#include "attack.h"
#include "board.h"
#include "colour.h"
#include "eval.h"
#include "list.h"
#include "move.h"
#include "move_check.h"
#include "move_do.h"
#include "option.h"
#include "piece.h"
#include "pst.h"
#include "pv.h"
#include "recog.h"
#include "search.h"
#include "search_full.h"
#include "see.h"
#include "sort.h"
#include "trans.h"
#include "util.h"
#include "value.h"

static const bool UseTrans = true;
static const int TransDepth = 1;

static const int NodeAll = -1;
static const int NodePV  =  0;
static const int NodeCut = +1;

#define NODE_OPP(type)     (-(type))
#define DEPTH_MATCH(d1,d2) ((d1)>=(d2))

static int  full_root            (list_t * list, board_t * board, int alpha, int beta, int depth, int height, int search_type);
static int  full_search          (board_t * board, int alpha, int beta, int depth, int height, mv_t pv[], int node_type);
static int  full_no_null         (board_t * board, int alpha, int beta, int depth, int height, mv_t pv[], int node_type, int trans_move, int * best_move);
static int  full_quiescence      (board_t * board, int alpha, int beta, int depth, int height, mv_t pv[]);
static void pv_fill              (const mv_t pv[], board_t * board);
static bool simple_stalemate     (const board_t * board);

void search_full_init(list_t * list, board_t * board) {
   const char * string;
   int trans_move, trans_min_depth, trans_max_depth, trans_min_value, trans_max_value;

   list_note(list);
   list_sort(list);
   trans_move = MoveNone;
   if (UseTrans) trans_retrieve(Trans,board->key,&trans_move,&trans_min_depth,&trans_max_depth,&trans_min_value,&trans_max_value);
   note_moves(list,board,0,trans_move);
   list_sort(list);
}

int search_full_root(list_t * list, board_t * board, int depth, int search_type) {
   int value;

   value = full_root(list,board,-ValueInf,+ValueInf,depth,0,search_type);
   return value;
}

static int full_root(list_t * list, board_t * board, int alpha, int beta, int depth, int height, int search_type) {
   int old_alpha;
   int value, best_value;
   int i, move;
   int new_depth;
   undo_t undo[1];
   mv_t new_pv[HeightMax];

   SearchCurrent->node_nb++;
   SearchInfo->check_nb--;
   for (i = 0; i < LIST_SIZE(list); i++) list->value[i] = ValueNone;
   old_alpha = alpha;
   best_value = ValueNone;
   for (i = 0; i < LIST_SIZE(list); i++) {
      move = LIST_MOVE(list,i);
      SearchRoot->depth = depth;
      SearchRoot->move = move;
      SearchRoot->move_pos = i;
      SearchRoot->move_nb = LIST_SIZE(list);
      search_update_root();
      new_depth = depth-1;
      move_do(board,move,undo);
      if (best_value == ValueNone) { // first move
         value = -full_search(board,-beta,-alpha,new_depth,height+1,new_pv,NodePV);
      } else { // other moves
         value = -full_search(board,-alpha-1,-alpha,new_depth,height+1,new_pv,NodeCut);
         if (value > alpha) { // && value < beta
            SearchRoot->change = true;
            SearchRoot->easy = false;
            SearchRoot->flag = false;
            search_update_root();
            value = -full_search(board,-beta,-alpha,new_depth,height+1,new_pv,NodePV);
         }
      }
      move_undo(board,move,undo);
      if (value <= alpha) { // upper bound
         list->value[i] = old_alpha;
      } else if (value >= beta) { // lower bound
         list->value[i] = beta;
      } else { // alpha < value < beta => exact value
         list->value[i] = value;
      }
      if (value > best_value && (best_value == ValueNone || value > alpha)) {
         SearchBest->move = move;
         SearchBest->value = value;
         if (value <= alpha) { // upper bound
            SearchBest->flags = SearchUpper;
         } else if (value >= beta) { // lower bound
            SearchBest->flags = SearchLower;
         } else { // alpha < value < beta => exact value
            SearchBest->flags = SearchExact;
         }
         SearchBest->depth = depth;
         pv_cat(SearchBest->pv,new_pv,move);
         search_update_best();
      }
      if (value > best_value) {
         best_value = value;
         if (value > alpha) {
            if (search_type == SearchNormal) alpha = value;
            if (value >= beta) break;
         }
      }
   }
   list_sort(list);
   if (UseTrans && best_value > old_alpha && best_value < beta) {
      pv_fill(SearchBest->pv,board);
   }
   return best_value;
}

static int full_search(board_t * board, int alpha, int beta, int depth, int height, mv_t pv[], int node_type) {
   bool in_check;
   bool single_reply;
   int trans_move, trans_depth, trans_min_depth, trans_max_depth, trans_min_value, trans_max_value;
   int min_value, max_value;
   int old_alpha;
   int value, best_value;
   int move, best_move;
   int new_depth;
   int played_nb;
   int i;
   attack_t attack[1];
   sort_t sort[1];
   undo_t undo[1];
   mv_t new_pv[HeightMax];
   mv_t played[256];

   if (depth <= 0) return full_quiescence(board,alpha,beta,0,height,pv);
   SearchCurrent->node_nb++;
   SearchInfo->check_nb--;
   PV_CLEAR(pv);
   if (height > SearchCurrent->max_depth) SearchCurrent->max_depth = height;
   if (SearchInfo->check_nb <= 0) {
      SearchInfo->check_nb += SearchInfo->check_inc;
      search_check();
   }
   if (board_is_repetition(board) || recog_draw(board)) return ValueDraw;
   trans_move = MoveNone;
   if (UseTrans && depth >= TransDepth) {
      if (trans_retrieve(Trans,board->key,&trans_move,&trans_min_depth,&trans_max_depth,&trans_min_value,&trans_max_value)) {
         if (node_type != NodePV) {
            min_value = -ValueInf;
            if (DEPTH_MATCH(trans_min_depth,depth)) {
               min_value = value_from_trans(trans_min_value,height);
               if (min_value >= beta) return min_value;
            }
            max_value = +ValueInf;
            if (DEPTH_MATCH(trans_max_depth,depth)) {
               max_value = value_from_trans(trans_max_value,height);
               if (max_value <= alpha) return max_value;
            }
            if (min_value == max_value) return min_value; // exact match
         }
      }
   }
   if (height >= HeightMax-1) return eval(board);
   old_alpha = alpha;
   best_value = ValueNone;
   best_move = MoveNone;
   played_nb = 0;
   attack_set(attack,board);
   in_check = ATTACK_IN_CHECK(attack);
   sort_init(sort,board,attack,depth,height,trans_move);
   single_reply = false;
   if (in_check && LIST_SIZE(sort->list) == 1) single_reply = true; // HACK
   while ((move=sort_next(sort)) != MoveNone) {
      new_depth = depth-1;
      move_do(board,move,undo);
      if (node_type != NodePV || best_value == ValueNone) { // first move
         value = -full_search(board,-beta,-alpha,new_depth,height+1,new_pv,NODE_OPP(node_type));
      } else { // other moves
         value = -full_search(board,-alpha-1,-alpha,new_depth,height+1,new_pv,NodeCut);
         if (value > alpha) { // && value < beta
            value = -full_search(board,-beta,-alpha,new_depth,height+1,new_pv,NodePV);
         }
      }
      move_undo(board,move,undo);
      played[played_nb++] = move;
      if (value > best_value) {
         best_value = value;
         pv_cat(pv,new_pv,move);
         if (value > alpha) {
            alpha = value;
            best_move = move;
            if (value >= beta) goto cut;
         }
      }
      if (node_type == NodeCut) node_type = NodeAll;
   }
   if (best_value == ValueNone) { // no legal move
      if (in_check) {
         return VALUE_MATE(height);
      } else {
         return ValueDraw;
      }
   }
cut:
   if (best_move != MoveNone) {
      good_move(best_move,board,depth,height);
      if (best_value >= beta && !move_is_tactical(best_move,board)) {
         for (i = 0; i < played_nb-1; i++) {
            move = played[i];
            history_bad(move,board);
         }
         history_good(best_move,board);
      }
   }
   if (UseTrans && depth >= TransDepth) {
      trans_move = best_move;
      trans_depth = depth;
      trans_min_value = (best_value > old_alpha) ? value_to_trans(best_value,height) : -ValueInf;
      trans_max_value = (best_value < beta)      ? value_to_trans(best_value,height) : +ValueInf;
      trans_store(Trans,board->key,trans_move,trans_depth,trans_min_value,trans_max_value);
   }
   return best_value;
}

static int full_no_null(board_t * board, int alpha, int beta, int depth, int height, mv_t pv[], int node_type, int trans_move, int * best_move) {
   int value, best_value;
   int move;
   int new_depth;
   attack_t attack[1];
   sort_t sort[1];
   undo_t undo[1];
   mv_t new_pv[HeightMax];

   SearchCurrent->node_nb++;
   SearchInfo->check_nb--;
   PV_CLEAR(pv);
   if (height > SearchCurrent->max_depth) SearchCurrent->max_depth = height;
   if (SearchInfo->check_nb <= 0) {
      SearchInfo->check_nb += SearchInfo->check_inc;
      search_check();
   }
   attack_set(attack,board);
   *best_move = MoveNone;
   best_value = ValueNone;
   sort_init(sort,board,attack,depth,height,trans_move);
   while ((move=sort_next(sort)) != MoveNone) {
      new_depth = depth-1;
      move_do(board,move,undo);
      value = -full_search(board,-beta,-alpha,new_depth,height+1,new_pv,NODE_OPP(node_type));
      move_undo(board,move,undo);
      if (value > best_value) {
         best_value = value;
         pv_cat(pv,new_pv,move);
         if (value > alpha) {
            alpha = value;
            *best_move = move;
            if (value >= beta) goto cut;
         }
      }
   }
   if (best_value == ValueNone) { // no legal move => stalemate
      best_value = ValueDraw;
   }
cut:
   return best_value;
}

static int full_quiescence(board_t * board, int alpha, int beta, int depth, int height, mv_t pv[]) {
   bool in_check;
   int old_alpha;
   int value, best_value;
   int best_move;
   int move;
   attack_t attack[1];
   sort_t sort[1];
   undo_t undo[1];
   mv_t new_pv[HeightMax];

   SearchCurrent->node_nb++;
   SearchInfo->check_nb--;
   PV_CLEAR(pv);
   if (height > SearchCurrent->max_depth) SearchCurrent->max_depth = height;
   if (SearchInfo->check_nb <= 0) {
      SearchInfo->check_nb += SearchInfo->check_inc;
      search_check();
   }
   if (board_is_repetition(board) || recog_draw(board)) return ValueDraw;
   attack_set(attack,board);
   in_check = ATTACK_IN_CHECK(attack);
   if (height >= HeightMax-1) return eval(board);
   old_alpha = alpha;
   best_value = ValueNone;
   best_move = MoveNone;
   if (!in_check) {
      if (simple_stalemate(board)) return ValueDraw;
      value = eval(board);
      best_value = value;
      if (value > alpha) {
         alpha = value;
         if (value >= beta) goto cut;
      }
   }
   sort_init_qs(sort,board,attack,false);
   while ((move=sort_next_qs(sort)) != MoveNone) {
      move_do(board,move,undo);
      value = -full_quiescence(board,-beta,-alpha,depth-1,height+1,new_pv);
      move_undo(board,move,undo);
      if (value > best_value) {
         best_value = value;
         pv_cat(pv,new_pv,move);
         if (value > alpha) {
            alpha = value;
            best_move = move;
            if (value >= beta) goto cut;
         }
      }
   }
   if (best_value == ValueNone) { // no legal move
      return VALUE_MATE(height);
   }
cut:
   return best_value;
}

static void pv_fill(const mv_t pv[], board_t * board) {
   int move;
   int trans_move, trans_depth, trans_min_value, trans_max_value;
   undo_t undo[1];

   move = *pv;
   if (move != MoveNone && move != MoveNull) {
      move_do(board,move,undo);
      pv_fill(pv+1,board);
      move_undo(board,move,undo);
      trans_move = move;
      trans_depth = -127; // HACK
      trans_min_value = -ValueInf;
      trans_max_value = +ValueInf;
      trans_store(Trans,board->key,trans_move,trans_depth,trans_min_value,trans_max_value);
   }
}

static bool simple_stalemate(const board_t * board) {
   int me, opp;
   int king;
   int opp_flag;
   int from, to;
   int capture;
   const inc_t * inc_ptr;
   int inc;

   me = board->turn;
   if (board->piece_size[me] != 1 || board->pawn_size[me] != 0) return false; // no
   king = KING_POS(board,me);
   if (king != A1 && king != H1 && king != A8 && king != H8) return false; // no
   opp = COLOUR_OPP(me);
   opp_flag = COLOUR_FLAG(opp);
   from = king;
   for (inc_ptr = KingInc; (inc=*inc_ptr) != IncNone; inc_ptr++) {
      to = from + inc;
      capture = board->square[to];
      if (capture == Empty || FLAG_IS(capture,opp_flag)) {
         if (!is_attacked(board,to,opp)) return false; // legal king move
      }
   }
   return true;
}
